# Stolen from https://github.com/falconry/falcon/blob/master/falcon/status_codes.py
HTTP_100 = '100 Continue'
HTTP_CONTINUE = HTTP_100
HTTP_101 = '101 Switching Protocols'
HTTP_SWITCHING_PROTOCOLS = HTTP_101

HTTP_200 = '200 OK'
HTTP_OK = HTTP_200
HTTP_201 = '201 Created'
HTTP_CREATED = HTTP_201
HTTP_202 = '202 Accepted'
HTTP_ACCEPTED = HTTP_202
HTTP_203 = '203 Non-Authoritative Information'
HTTP_NON_AUTHORITATIVE_INFORMATION = HTTP_203
HTTP_204 = '204 No Content'
HTTP_NO_CONTENT = HTTP_204
HTTP_205 = '205 Reset Content'
HTTP_RESET_CONTENT = HTTP_205
HTTP_206 = '206 Partial Content'
HTTP_PARTIAL_CONTENT = HTTP_206
HTTP_226 = '226 IM Used'
HTTP_IM_USED = HTTP_226

HTTP_300 = '300 Multiple Choices'
HTTP_MULTIPLE_CHOICES = HTTP_300
HTTP_301 = '301 Moved Permanently'
HTTP_MOVED_PERMANENTLY = HTTP_301
HTTP_302 = '302 Found'
HTTP_FOUND = HTTP_302
HTTP_303 = '303 See Other'
HTTP_SEE_OTHER = HTTP_303
HTTP_304 = '304 Not Modified'
HTTP_NOT_MODIFIED = HTTP_304
HTTP_305 = '305 Use Proxy'
HTTP_USE_PROXY = HTTP_305
HTTP_307 = '307 Temporary Redirect'
HTTP_TEMPORARY_REDIRECT = HTTP_307
HTTP_308 = '308 Permanent Redirect'
HTTP_PERMANENT_REDIRECT = HTTP_308

HTTP_400 = '400 Bad Request'
HTTP_BAD_REQUEST = HTTP_400
HTTP_401 = '401 Unauthorized'  # <-- Really means "unauthenticated"
HTTP_UNAUTHORIZED = HTTP_401
HTTP_402 = '402 Payment Required'
HTTP_PAYMENT_REQUIRED = HTTP_402
HTTP_403 = '403 Forbidden'  # <-- Really means "unauthorized"
HTTP_FORBIDDEN = HTTP_403
HTTP_404 = '404 Not Found'
HTTP_NOT_FOUND = HTTP_404
HTTP_405 = '405 Method Not Allowed'
HTTP_METHOD_NOT_ALLOWED = HTTP_405
HTTP_406 = '406 Not Acceptable'
HTTP_NOT_ACCEPTABLE = HTTP_406
HTTP_407 = '407 Proxy Authentication Required'
HTTP_PROXY_AUTHENTICATION_REQUIRED = HTTP_407
HTTP_408 = '408 Request Time-out'
HTTP_REQUEST_TIMEOUT = HTTP_408
HTTP_409 = '409 Conflict'
HTTP_CONFLICT = HTTP_409
HTTP_410 = '410 Gone'
HTTP_GONE = HTTP_410
HTTP_411 = '411 Length Required'
HTTP_LENGTH_REQUIRED = HTTP_411
HTTP_412 = '412 Precondition Failed'
HTTP_PRECONDITION_FAILED = HTTP_412
HTTP_413 = '413 Payload Too Large'
HTTP_REQUEST_ENTITY_TOO_LARGE = HTTP_413
HTTP_414 = '414 URI Too Long'
HTTP_REQUEST_URI_TOO_LONG = HTTP_414
HTTP_415 = '415 Unsupported Media Type'
HTTP_UNSUPPORTED_MEDIA_TYPE = HTTP_415
HTTP_416 = '416 Range Not Satisfiable'
HTTP_REQUESTED_RANGE_NOT_SATISFIABLE = HTTP_416
HTTP_417 = '417 Expectation Failed'
HTTP_EXPECTATION_FAILED = HTTP_417
HTTP_418 = "418 I'm a teapot"
HTTP_IM_A_TEAPOT = HTTP_418
HTTP_422 = '422 Unprocessable Entity'
HTTP_UNPROCESSABLE_ENTITY = HTTP_422
HTTP_426 = '426 Upgrade Required'
HTTP_UPGRADE_REQUIRED = HTTP_426
HTTP_428 = '428 Precondition Required'
HTTP_PRECONDITION_REQUIRED = HTTP_428
HTTP_429 = '429 Too Many Requests'
HTTP_TOO_MANY_REQUESTS = HTTP_429
HTTP_431 = '431 Request Header Fields Too Large'
HTTP_REQUEST_HEADER_FIELDS_TOO_LARGE = HTTP_431
HTTP_451 = '451 Unavailable For Legal Reasons'
HTTP_UNAVAILABLE_FOR_LEGAL_REASONS = HTTP_451

HTTP_500 = '500 Internal Server Error'
HTTP_INTERNAL_SERVER_ERROR = HTTP_500
HTTP_501 = '501 Not Implemented'
HTTP_NOT_IMPLEMENTED = HTTP_501
HTTP_502 = '502 Bad Gateway'
HTTP_BAD_GATEWAY = HTTP_502
HTTP_503 = '503 Service Unavailable'
HTTP_SERVICE_UNAVAILABLE = HTTP_503
HTTP_504 = '504 Gateway Time-out'
HTTP_GATEWAY_TIMEOUT = HTTP_504
HTTP_505 = '505 HTTP Version not supported'
HTTP_HTTP_VERSION_NOT_SUPPORTED = HTTP_505
HTTP_511 = '511 Network Authentication Required'
HTTP_NETWORK_AUTHENTICATION_REQUIRED = HTTP_511

# 70X - Inexcusable
HTTP_701 = '701 Meh'
HTTP_702 = '702 Emacs'
HTTP_703 = '703 Explosion'

# 71X - Novelty Implementations
HTTP_710 = '710 PHP'
HTTP_711 = '711 Convenience Store'
HTTP_712 = '712 NoSQL'
HTTP_719 = '719 I am not a teapot'

# 72X - Edge Cases
HTTP_720 = '720 Unpossible'
HTTP_721 = '721 Known Unknowns'
HTTP_722 = '722 Unknown Unknowns'
HTTP_723 = '723 Tricky'
HTTP_724 = '724 This line should be unreachable'
HTTP_725 = '725 It works on my machine'
HTTP_726 = "726 It's a feature, not a bug"
HTTP_727 = '727 32 bits is plenty'

# 74X - Meme Driven
HTTP_740 = '740 Computer says no'
HTTP_741 = '741 Compiling'
HTTP_742 = '742 A kitten dies'
HTTP_743 = '743 I thought I knew regular expressions'
HTTP_744 = '744 Y U NO write integration tests?'
HTTP_745 = ("745 I don't always test my code, but when I do"
            'I do it in production')
HTTP_748 = '748 Confounded by Ponies'
HTTP_749 = '749 Reserved for Chuck Norris'

# 75X - Syntax Errors
HTTP_750 = "750 Didn't bother to compile it"
HTTP_753 = '753 Syntax Error'
HTTP_754 = '754 Too many semi-colons'
HTTP_755 = '755 Not enough semi-colons'
HTTP_759 = '759 Unexpected T_PAAMAYIM_NEKUDOTAYIM'

# 77X - Predictable Problems
HTTP_771 = '771 Cached for too long'
HTTP_772 = '772 Not cached long enough'
HTTP_773 = '773 Not cached at all'
HTTP_774 = '774 Why was this cached?'
HTTP_776 = '776 Error on the Exception'
HTTP_777 = '777 Coincidence'
HTTP_778 = '778 Off By One Error'
HTTP_779 = '779 Off By Too Many To Count Error'

# 78X - Somebody Else's Problem
HTTP_780 = '780 Project owner not responding'
HTTP_781 = '781 Operations'
HTTP_782 = '782 QA'
HTTP_783 = '783 It was a customer request, honestly'
HTTP_784 = '784 Management, obviously'
HTTP_785 = '785 TPS Cover Sheet not attached'
HTTP_786 = '786 Try it now'

# 79X - Internet crashed
HTTP_791 = '791 The Internet shut down due to copyright restrictions'
HTTP_792 = '792 Climate change driven catastrophic weather event'
HTTP_797 = '797 This is the last page of the Internet. Go back'
HTTP_799 = '799 End of the world'


import re
import asyncio
import urllib.parse as parse
from datetime import datetime
from albatross import Request, Response
from albatross.status_codes import HTTP_404, HTTP_500, HTTP_405
from albatross.http_error import HTTPError
import traceback


class Server:
    """The core albatross server

    Attributes:
        _handlers (list): a list of route-handler tuples
        _middleware (list): a list of middlewares to process requests
    """
    def __init__(self):
        self._handlers = []
        self._middleware = []
        self.max_read_chunk = 1024*1024
        self.debug = False
        self.spoof_options = True

    def get_handler(self, path):
        for route, handler in self._handlers:
            match = route.match(path)
            if match:
                return handler, match.groupdict()
        return None, None

    def add_route(self, route, handler):
        route += '$'
        compiled = re.compile(route)
        self._handlers.append((compiled, handler))

    def add_middleware(self, middleware):
        self._middleware.append(middleware)

    def _parse_header_lines(self, lines):
        headers = {}
        for l in lines:
            try:
                key, value = l.split(':', 1)
                headers[key] = value.strip()
            except ValueError:
                pass
        return headers

    async def _parse_request(self, request_reader):
        request_line = await request_reader.readline()
        request_line = request_line.decode()
        method, url_string, _ = request_line.split(' ', 2)
        method = method.upper()
        url = parse.urlparse(url_string)

        header_lines = []
        while True:
            l = await request_reader.readline()
            if l == b'\r\n':
                break
            header_lines.append(l.decode())

        headers = self._parse_header_lines(header_lines)

        body = None
        if method in {'POST', 'PUT'}:
            body_parts = []
            content_length = int(headers.get('Content-Length', 0))
            while content_length > 0:
                read_chunk = max(content_length, self.max_read_chunk)
                body = await request_reader.read(read_chunk)
                body_parts.append(body.decode())
                content_length -= read_chunk
            body = ''.join(body_parts)

        path = url.path
        query = url.query

        handler, args = self.get_handler(path)

        req = Request(method, path, query, body, args, headers)

        return req, handler

    async def _route_request(self, handler, req, res):
        method = req.method
        if handler is None:
            raise HTTPError(HTTP_404)
        elif method == 'GET' and hasattr(handler, 'on_get'):
            await handler.on_get(req, res)
        elif method == 'POST' and hasattr(handler, 'on_post'):
            await handler.on_post(req, res)
        elif method == 'PUT' and hasattr(handler, 'on_put'):
            await handler.on_put(req, res)
        elif method == 'DELETE' and hasattr(handler, 'on_delete'):
            await handler.on_delete(req, res)
        elif method == 'OPTIONS':
            if hasattr(handler, 'on_options'):
                await handler.on_options(req, res)
            elif self.spoof_options:
                res.headers['Allow'] = 'GET,POST,DELETE,PUT'
        else:
            raise HTTPError(HTTP_405)

    async def _handle(self, request_reader, response_writer):
        """Takes reader and writer from asyncio loop server and writes the response to the request.

        :param request_reader:
        :param response_writer:
        :return:
        """
        req, handler = await self._parse_request(request_reader)

        res = Response()

        try:
            for middleware in self._middleware:
                await middleware.process_request(req, res, handler)

            await self._route_request(handler, req, res)
        except Exception as e:
            res.clear()
            if isinstance(e, HTTPError):
                res.status_code = e.status_code
            else:
                res.status_code = HTTP_500
            self.handle_error(res, e)

        for middleware in self._middleware:
            await middleware.process_response(req, res, handler)

        self._write_response(res, response_writer)
        await response_writer.drain()
        response_writer.close()

    def handle_error(self, res, e):
        res.write(res.status_code)
        traceback.print_exc()

    def _write_response(self, res, writer):
        writer.write(b'HTTP/1.0 %s\r\n' % res.status_code.encode())
        for key, value in res.headers.items():
            writer.write(key.encode() + b': ' + value.encode() + b'\r\n')
        for key, value in res.cookies.items():
            if isinstance(value, tuple):
                value, duration = value
                if isinstance(duration, datetime):
                    duration = duration.strftime('%a %d %b %Y %H:%M:%S GMT')
                    writer.write(b'Set-Cookie: %s=%s;expires=%s\r\n' % (key.encode(), value.encode(), duration))
                elif isinstance(duration, int):
                    writer.write(b'Set-Cookie: %s=%s;max-age=%s\r\n' % (key.encode(), value.encode(), duration))
            else:
                writer.write(b'Set-Cookie: %s=%s\r\n' % (key.encode(), value.encode()))
        writer.write(b'\r\n')
        for chunk in res._chunks:
            writer.write(chunk)
        writer.write_eof()

    async def initialize(self):
        pass

    def serve(self, port=8000, host='0.0.0.0', debug=False):
        if debug:
            self.debug = True

        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.initialize())
        print('Serving on %s:%d' % (host, port))
        loop.create_task(asyncio.start_server(self._handle, host, port))
        loop.run_forever()
        loop.close()



import urllib.parse as parse
import ujson as json
import re


class Request:
    """
    Attributes:
        method (str): One of GET, POST, PUT, DELETE
        path (str): Full request path
        query_string (str): Full query string
        query (dict): dict of the request query
        body (str): Request body
        args (dict): Dictionary of named parameters in route regex
        form (dict): Dictionary of body parameters
    """
    def __init__(self, method, path, query_string, body, args, headers):
        self.method = method
        self.path = path
        self.query_string = query_string
        self.body = body
        self.query = parse.parse_qs(query_string)
        self.args = args
        self.headers = headers
        self.cookies = {}
        self.form = None
        if body:
            self._parse_body()
        if 'Cookie' in headers:
            self.cookies = self._parse_cookie(headers['Cookie'])

    def _parse_cookie(self, value):
        cookie_pairs = re.findall('(\w+)=([^,;]+)', value)
        return dict(cookie_pairs)

    def _parse_form(self):
        content_type = self.headers['Content-Type']
        marker = 'boundary='
        index = content_type.find(marker)
        boundary = '--' + content_type[index + len(marker):]
        values = {}
        for group in re.findall('name=\"([a-zA-Z0-9-]+)\"\r\n\r\n(.*?)\r\n' + boundary, self.body,  re.MULTILINE):
            values[group[0]] = group[1]
        return values

    def _parse_body(self):
        content_type = self.headers['Content-Type']
        if content_type == 'application/x-www-form-urlencoded':
            self.form = parse.parse_qs(self.body)
        elif content_type == 'application/json':
            self.form = json.loads(self.body)
        elif content_type.startswith('multipart/form-data'):
            self.form = self._parse_form()



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='albatross3',
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.6',
    description='A modern async python3 web framework',
    long_description=long_description,
    url='https://github.com/kespindler/albatross',
    author='Kurt Spindler',
    author_email='kespindler@gmail.com',
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='web http server async',
    packages=['albatross'],
    install_requires=[],
    extras_require={
        'dev': [],
        'test': [],
    },
)
